#pragma once
#include "COException.h"

//singleton class for property lookups

class PropertyMap
{
public:

	class ConstantValue {
		bool isReal;
		double realValue;
		CBStr stringValue;
		public:
		ConstantValue() {
		}
		ConstantValue(double real) {
			isReal=true;
			realValue=real;
		}
		ConstantValue(const wchar_t *string) {
			isReal=false;
			stringValue=string;
		}
		ConstantValue(const ConstantValue &other) {
			isReal=other.isReal;
			realValue=other.realValue;
			stringValue=other.stringValue;
		}
		const ConstantValue &operator=(const ConstantValue &other) {
			isReal=other.isReal;
			realValue=other.realValue;
			stringValue=other.stringValue;
			return *this;
		}
		void GetVariant(CVariant &res) const {
			if (isReal) {
				res.Set(realValue);
			} else {
				res.Set(stringValue);
			}
		}
	};

	enum TemperatureDependentProperty {
		SurfaceTensionSatLiquid,
		ThermalConductivityOfLiquid,
		ThermalConductivityOfVapor,
		VaporPressure,
		VolumeChangeUponVaporization,
		VolumeOfLiquid,
		ViscosityOfLiquid,
		ViscosityOfVapor,
		IdealGasEnthalpy,
		IdealGasEntropy
	};

	enum SinglePhaseProperty {
		Density,
		Enthalpy,
		Entropy,
		GibbsEnergy,
		HeatCapacityCp,
		HeatCapacityCv,
		InternalEnergy,
		MolecularWeight,
		ThermalConductivity,
		Volume,
		Viscosity
	};

	enum TwoPhaseProperty {
		SurfaceTension
	};

private:

	StringMap<ConstantValue> constantMap; //for this package, we can simply put the constant values directly in a map
	CVariant constPropList;
	StringMap<TemperatureDependentProperty> tDepPropMap;
	CVariant tDepPropList;
	StringMap<SinglePhaseProperty> singlePhasePropMap;
	CVariant singlePhasePropList;
	CVariant twoPhasePropList;
	StringMap<double> universalConstantMap;
	CVariant universalConstantList;

public:
	
	PropertyMap() {
		int index;
		//compound constants
		constantMap[L"criticalDensity"]=ConstantValue(Water::RHOCRIT/(Water::MOLWT*1e-3));
		constantMap[L"criticalPressure"]=ConstantValue(Water::PCRIT*1e6);
		constantMap[L"criticalTemperature"]=ConstantValue(Water::TCRIT);
		constantMap[L"criticalVolume"]=ConstantValue(Water::MOLWT*1e-3/Water::RHOCRIT);
		constantMap[L"molecularWeight"]=ConstantValue(Water::MOLWT);
		constantMap[L"triplePointPressure"]=ConstantValue(Water::PTRIP*1e6);
		constantMap[L"triplePointTemperature"]=ConstantValue(Water::TTRIP);
		constantMap[L"casRegistryNumber"]=ConstantValue(L"7732-18-5");
		constantMap[L"chemicalFormula"]=ConstantValue(L"H2O");
		constantMap[L"iupacName"]=ConstantValue(L"oxidane");
		constantMap[L"normalBoilingPoint"]=ConstantValue(Water::Tsat(101325.0e-6));
		//compound constant list
		constPropList.MakeArray(constantMap.Count(),VT_BSTR);
		index=0;
		constantMap.Operate([this,&index](const wstring &key,const ConstantValue &value) -> bool {
			constPropList.SetStringAt(index++,CBStr(key));
			return true;
		});
		//temperature dependent property map
		tDepPropMap[L"SurfaceTensionSatLiquid"]=SurfaceTensionSatLiquid;
		tDepPropMap[L"ThermalConductivityOfLiquid"]=ThermalConductivityOfLiquid;
		tDepPropMap[L"ThermalConductivityOfVapor"]=ThermalConductivityOfVapor;
		tDepPropMap[L"VaporPressure"]=VaporPressure;
		tDepPropMap[L"VolumeChangeUponVaporization"]=VolumeChangeUponVaporization;
		tDepPropMap[L"VolumeOfLiquid"]=VolumeOfLiquid;
		tDepPropMap[L"ViscosityOfLiquid"]=ViscosityOfLiquid;
		tDepPropMap[L"ViscosityOfVapor"]=ViscosityOfVapor;
		tDepPropMap[L"IdealGasEnthalpy"]=IdealGasEnthalpy;
		tDepPropMap[L"IdealGasEntropy"]=IdealGasEntropy;
		//temperature dependent property list
		tDepPropList.MakeArray(tDepPropMap.Count(),VT_BSTR);
		index=0;
		tDepPropMap.Operate([this,&index](const wstring &key,const TemperatureDependentProperty value) -> bool {
			tDepPropList.SetStringAt(index++,CBStr(key));
			return true;
		});
		//single phase properties
		singlePhasePropMap[L"Density"]=Density;
		singlePhasePropMap[L"Enthalpy"]=Enthalpy;
		singlePhasePropMap[L"Entropy"]=Entropy;
		singlePhasePropMap[L"GibbsEnergy"]=GibbsEnergy;
		singlePhasePropMap[L"HeatCapacityCp"]=HeatCapacityCp;
		singlePhasePropMap[L"HeatCapacityCv"]=HeatCapacityCv;
		singlePhasePropMap[L"InternalEnergy"]=InternalEnergy;
		singlePhasePropMap[L"MolecularWeight"]=MolecularWeight;
		singlePhasePropMap[L"ThermalConductivity"]=ThermalConductivity;
		singlePhasePropMap[L"Volume"]=Volume;
		singlePhasePropMap[L"Viscosity"]=Viscosity;
		//single phase property list
		singlePhasePropList.MakeArray(singlePhasePropMap.Count(),VT_BSTR);
		index=0;
		singlePhasePropMap.Operate([this,&index](const wstring &key,const SinglePhaseProperty value) -> bool {
			singlePhasePropList.SetStringAt(index++,CBStr(key));
			return true;
		});
		//two phase property list (we do not implement a lookup map as we have only one)
		twoPhasePropList.MakeArray(1,VT_BSTR);
		twoPhasePropList.SetStringAt(0,CBStr(L"SurfaceTension"));
		//universal constants
		universalConstantMap[L"molarGasConstant"]=8.31447215;
		universalConstantMap[L"avogadroConstant"]=6.0221419947e23;
		universalConstantMap[L"boltzmannConstant"]=1.380650324e-23;
		universalConstantMap[L"speedOfLightInVacuum"]=299792458.1;
		universalConstantMap[L"standardAccelerationOfGravity"]=9.80665;
		//single phase property list
		universalConstantList.MakeArray(universalConstantMap.Count(),VT_BSTR);
		index=0;
		universalConstantMap.Operate([this,&index](const wstring &key,const double value) -> bool {
			universalConstantList.SetStringAt(index++,CBStr(key));
			return true;
		});
	}

	void GetConstValue(const wchar_t *constName,CVariant &res) const {
		if (!constName) throw COException(L"Empty compound constant name",ECapeInvalidArgumentHR);
		const ConstantValue *val;
		if (!constantMap.LookupRef(constName,val)) {
			wstring error;
			error=L"Invalid/unsupported compound constant: ";
			error+=constName;
			throw COException(error,ECapeInvalidArgumentHR);
		}
		val->GetVariant(res);		
	}

	CVariant GetConstList() const {
		//return a copy
		CVariant res(constPropList.Copy(),true);
		return res;
	}

	TemperatureDependentProperty GetTDepProperty(const wchar_t *name) const {
		if (!name) throw COException(L"Empty temperature dependent property name",ECapeInvalidArgumentHR);
		TemperatureDependentProperty prop;
		if (!tDepPropMap.Lookup(name,prop)) {
			wstring error;
			error=L"Invalid/unsupported temperature dependent property: ";
			error+=name;
			throw COException(error,ECapeInvalidArgumentHR);
		}
		return prop;
	}

	CVariant GetTDepPropList() const {
		//return a copy
		CVariant res(tDepPropList.Copy(),true);
		return res;
	}

	SinglePhaseProperty GetSinglePhaseProperty(const wchar_t *name) const {
		if (!name) throw COException(L"Empty single phase property name",ECapeInvalidArgumentHR);
		SinglePhaseProperty prop;
		if (!singlePhasePropMap.Lookup(name,prop)) {
			wstring error;
			error=L"Invalid/unsupported single phase property: ";
			error+=name;
			throw COException(error,ECapeInvalidArgumentHR);
		}
		return prop;
	}

	CVariant GetSinglePhasePropList() const {
		//return a copy
		CVariant res(singlePhasePropList.Copy(),true);
		return res;
	}

	TwoPhaseProperty GetTwoPhaseProperty(const wchar_t *name) const {
		if (SameString(name,L"SurfaceTension")) return SurfaceTension;
		if (!name) throw COException(L"Empty two phase property name",ECapeInvalidArgumentHR);
		wstring error;
		error=L"Invalid/unsupported two phase property: ";
		error+=name;
		throw COException(error,ECapeInvalidArgumentHR);
	}

	CVariant GetTwoPhasePropList() const {
		//return a copy
		CVariant res(twoPhasePropList.Copy(),true);
		return res;
	}

	double GetUniversalConstant(const wchar_t *name) {
		if (!name) throw COException(L"Empty universal constant property name",ECapeInvalidArgumentHR);
		double value;
		if (!universalConstantMap.Lookup(name,value)) {
			wstring error;
			error=L"Invalid/unsupported universal constant: ";
			error+=name;
			throw COException(error,ECapeInvalidArgumentHR);
		}
		return value;
	}

	CVariant GetUniversalConstantList() const {
		//return a copy
		CVariant res(universalConstantList.Copy(),true);
		return res;
	}


};

//singleton
extern PropertyMap propertyMap;
